/*
 * Copyright 2004 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package moazs.api;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.xml.transform.TransformerException;

import moazs.Constants;
import moazs.external.MoaSZR;
import moazs.external.ZkopfRequest;
import moazs.modell.Identification;
import moazs.modell.Internetaddress;
import moazs.modell.Payload;
import moazs.modell.Postaladdress;
import moazs.modell.Rcpt;
import moazs.modell.Request;
import moazs.modell.Zuseserver;
import moazs.util.HibernateUtil;
import moazs.util.MoaConfig;
import moazs.util.MoaFactory;
import moazs.util.MoaXMLUtils;
import net.sf.workqueue.QueueException;
import net.sf.workqueue.WorkQueue;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.Transaction;
import org.hibernate.classic.Session;
import org.w3c.dom.Document;

import at.gv.egovernment.moa.util.DOMUtils;


public class MoaSync {

    static private Logger logger = Logger.getLogger(MoaSync.class);
    private static int count = 0;

    private SyncStatus syncStatus = null;
    private boolean running = false;
    
    private boolean stopped = false;
    
    private static Hashtable syncWatchDogThreads = new Hashtable();
    
    private MoaConfig config = null;
    
    public MoaSync() {
      config = new MoaConfig();
      syncStatus = new SyncStatus();
    }
   
    public synchronized SyncStatus getSyncStatus() {
      return this.syncStatus;
    }
    
    public boolean isStopped() {
      return this.stopped;
    }
    
    public void terminate() {
      this.running = false;
    }
    
    public static synchronized List getSyncWatchDogThreads() {
      Enumeration threadKeys = syncWatchDogThreads.keys();
      ArrayList resultList = new ArrayList();
      while (threadKeys.hasMoreElements()) {
        resultList.add(syncWatchDogThreads.get(threadKeys.nextElement()));
      }
      return resultList;
    }
    
    public static synchronized MoaSync getSyncWatchDogThread(String id) {
      return (MoaSync) syncWatchDogThreads.get(id);
    }
    
    public synchronized void removeSyncWatchDogThread(String id) throws MoaException {
      try {
        syncWatchDogThreads.remove(id);
      } catch(NullPointerException e) {
        throw new MoaException("Thread with id " + id + " not found in thread list.");
      }
    }
    
    public synchronized void addSyncWatchDogThread(String id, MoaSync watchDog) throws MoaException {
      Object o = syncWatchDogThreads.get(id);
      if (o != null) {
        throw new MoaException("Thread with given id already registered.");
      }
      logger.debug("Adding sync watchdog thread (" + id + ")");
      syncWatchDogThreads.put(id, watchDog);
    }
    
    private synchronized void inc() {
      count ++;
    }

    private synchronized void dec()  {
      count --; 
    }


    public static int getCount() {
      return count;
    }
 
    /**
    * Methode die einen DeliveryRequest engegen nimmt, die Daten auf vollstndigkeit Prft und die 
    * Requsts dann in die Queue einsortiert 
     * @param in Document
     * @return geprueftes Document
     */
    public Document DeliveryRequest(Document in) {

      // debug incoming documents
      if (config.getDebugRequests()) {
        try {
          logger.debug(MoaXMLUtils.serializeToString(in.getDocumentElement(), true));
        } catch (IOException e1) {
          logger.warn("Error writing incoming request.", e1);
        }
      }
      
      inc(); 
      WorkQueue queue = MoaFactory.getQueue(); 
      
      long start = System.currentTimeMillis();
      long requestId = -1l;
      int queueId   = -1;
      
      boolean partial = false;
      
      // sync request
      boolean sync = false;
      Request r = null;

      Session sess    = HibernateUtil.getCurrentSession();
      Transaction tx = sess.beginTransaction();
      boolean error = true;
        try {
          
          logger.debug("Performing schema validation: " + config.getSchemaValidation());
          
          if (config.getSchemaValidation()) {
            // check schema validation
            try {
              if (!MoaXMLUtils.validateMZS(in.getDocumentElement())) {
                throw new MoaException(Constants.ERROR_REQUEST_SCHEMA_VALIDATION_FAILED, "Request schema validation failed.");
              }
              logger.debug("Schema successfully validated.");
            } catch (Throwable e) {
              logger.debug("Schema validation failed.", e);
              throw new MoaException(Constants.ERROR_REQUEST_SCHEMA_VALIDATION_FAILED, "Request schema validation failed.");
            }
          }

          
          // Mappen 
          logger.debug("Saving request in database.");
          r = MoaMapper.saveRequest(sess, in);

          // check for sync delivery
          sync = r.getSyncId() != null;
          
          // MOA-Sync
          if (sync) {
            logger.info("Sync request detected.");
            running = true;
            addSyncWatchDogThread(r.getSyncId(), this);
          }
            
          Zuseserver zsvr = r.getZuseserver();
          
          if (zsvr == null) {
            // BEGIN ZUSEKOPF
            logger.debug("Starting delivery head query");
            // build list of mimetypes of documents that should be delivered
            List mimeTypes = new ArrayList();
            if (r.getXmlDocument() != null) {
              mimeTypes.add(r.getXmlDocument().getMimetype());
            }
            if (r.getPayload() != null) {
              Iterator it = r.getPayload().iterator();
              while (it.hasNext()) {
                mimeTypes.add(((Payload) it.next()).getMimetype());
              }
            }
            
            try {
              zsvr = getZuseserver(r.getRcpt(), r.getRequiresEncryption(), r.isPrivateDelivery(), r.getSenderProfileId(), mimeTypes);
            } catch(MoaException e) {
              // only if not found try dual delivery
              if (e.getErrorCode() == Constants.ERROR_ZKOPF_NOT_FOUND) {
                if (r.getDualDelivery() != null) {
                  logger.info("Dual delivery requested.");
                  String dualDeliveryServer = r.getDualDelivery().getDualDeliveryServer();
                  
                  // check for available dual delivery server in config
                  if (dualDeliveryServer == null) {
                    String firstDualDeliveryServer = config.getFirstDualDeliveryServer();
                    if (firstDualDeliveryServer == null) {
                      logger.info("No dual delivery server found. Neither in request, nor in configuration.");
                      throw new MoaException(Constants.ERROR_DUAL_NO_SERVER_AVAILABLE, "No dual delivery server found. Neither in request nor in configuration.");
                    }
                    logger.info("Taking some dual delivery server from configuration.");
                    dualDeliveryServer = firstDualDeliveryServer;
                  } else {
                    String dualDeliveryServerString = r.getDualDelivery().getDualDeliveryServer();
                    if (config.getDualDeliveryServer(dualDeliveryServerString) == null) {
                      // check if is url
                      try {
                        new URL(dualDeliveryServerString);
                        logger.info("Dual delivery server url found in request.");
                        dualDeliveryServer = dualDeliveryServerString;
                      } catch (MalformedURLException err) {
                        throw new MoaException(Constants.ERROR_DUAL_SERVER_NOT_FOUND, "Given dual delivery server with friendlyname " + dualDeliveryServerString + " not found in configuration");
                      }
                    } else {
                      logger.info("Dual delivery server found in configuration.");
                      dualDeliveryServer = config.getDualDeliveryServer(dualDeliveryServerString);
                    }
                  }
                  
                  if (dualDeliveryServer != null) {
                    logger.info("Dual delivery server url: " + dualDeliveryServer);
                    // ok, we have the dual delivery server url, proceed
                    zsvr = new Zuseserver();
                    zsvr.setBpk("");
                    zsvr.setUrl(dualDeliveryServer);
                    zsvr.setMimetypes("*/*");
                  }
                }

              } else {
                // rethrow exception
                throw e;
              }
            }

            if (zsvr == null) {
                logger.warn("No delivery server available.");
                if (tx.isActive()) {
                  tx.rollback();
                }
                throw new MoaException(Constants.ERROR_ZKOPF_NOT_FOUND, "Person unadressable.");
            }
            // END ZUSEKOPF
          } else {
            logger.info("Skipping delivery head query.");
          }
            
            
          sess.save(zsvr);
          r.setZuseserver(zsvr);
          queueId = MoaFactory.getMinload();

          if (queueId == MoaFactory.getQueueId() || getCount()>5) {

              logger.debug("Setting queue id.");
              r.setQueueId( queueId );
          } else {
              logger.debug("Sync: Irgendwer ...");
              r.setQueueId(-1);
          }
          // Persistieren
          requestId = r.getId();
          partial = MoaMapper.isPartialSuccess(r.getPayload());

          tx.commit();
          error = false;
          //sess.flush();
          //sess.connection().commit();
      } catch (HibernateException he) {
        logger.warn("HibernateException", he);
        return MoaResponse.getError(r.getAppDeliveryId(), String.valueOf(r.getId()), "500", "Hibernate Exception", null);
      } catch (MoaException me) {
          logger.warn("MoaException", me);
          String mzsId = "";
          String appDeliveryId = "";

          if (r != null) {
              appDeliveryId = r.getAppDeliveryId();
              mzsId = String.valueOf(r.getId());
          }
          return MoaResponse.getError(appDeliveryId, mzsId, String.valueOf(me.getErrorCode()), me.getMessage(), null);
      } catch (Exception e) {
          logger.warn("Exception", e);
          String deliverid = "";
          String errmsg = "MoaException";
          
          if (r != null) {
              deliverid = r.getAppDeliveryId();
          }

          errmsg = e.toString();
          return MoaResponse.getError(deliverid, "", "500", errmsg, null);
      } catch (Throwable t) {
        logger.warn("Throwable", t);
        return MoaResponse.getError(r.getAppDeliveryId(), String.valueOf(r.getId()), "500", "Internal Error", null);
      } finally {
          dec();
          if (error) {
            logger.debug("rollback...");
            try {
              if (HibernateUtil.getCurrentSession().getTransaction().isActive()) {
                HibernateUtil.getCurrentSession().getTransaction().rollback();
              }
            } catch(HibernateException err) {
              logger.warn("Could not rollback transaction.");
            }
          }
      }

      try {
          logger.debug("Sync: enqueue ("+ requestId +") ...");
          if (queueId == MoaFactory.getQueueId()) {
              queue.add( new Long( requestId ));
          }
      } catch (QueueException qe) {
        logger.warn("QueueException", qe);
  	    Session sess2 = HibernateUtil.getCurrentSession();
  	    tx = sess2.beginTransaction();
  	    error = true;
  	    try {
  		    Request req = (Request)sess2.load(Request.class, new Long(requestId));
  		    if (req == null) {
  		      throw new ObjectNotFoundException(new Long(requestId), Request.class.toString());
  		    }
  		    req.setQueueId( -1 );
  		    tx.commit();
  		    error = false;
  	    } catch (HibernateException he) {
  		    logger.warn("HibernateException", he);
  
  	    } finally {
  	      if (error) {
            logger.warn("rollback ...");
            try {
              if (HibernateUtil.getCurrentSession().getTransaction().isActive()) {
                HibernateUtil.getCurrentSession().getTransaction().rollback();
              }
            } catch(HibernateException err) {
              logger.warn("Could not rollback transaction.");
            }
          }
  	    }
      }
      
      long stop = System.currentTimeMillis();
      logger.debug("time: " + (stop-start) + "ms");
      
      
      // MOA-Sync (wait for result)
      if (sync) {

        int maxWaitTime = config.getIntProperty("MaxWaitTime", 10, "sync");
        //try {
        logger.debug("Waiting for sync thread " + r.getSyncId() + " to terminate (max wait time: " + maxWaitTime + " seconds)");
        long runningTime = 0;
        long INT = 50; // ms
        maxWaitTime = maxWaitTime * 1000;
        
        while (runningTime < maxWaitTime && syncStatus.getStatus() == SyncStatus.STATUS_UNKNOWN && running) {
          // do nothing but sleep
          try {
            Thread.sleep(INT);
            runningTime += INT;
          } catch (InterruptedException e) {
            logger.warn("InterruptedException:", e);
          }
        }
        if (runningTime >= maxWaitTime) {
          logger.debug("End of sync watchdog thread: " + r.getSyncId() + ". Wait time exceeded.");
        } else if (syncStatus.getStatus() != SyncStatus.STATUS_UNKNOWN) {
          logger.debug("End of sync watchdog thread: " + r.getSyncId() + ". Status changed to " + syncStatus.getStatus());
        } else {
          logger.debug("End of sync watchdog thread: " + r.getSyncId() + ". Thread terminated externally.");
        }        
        
        // wait time for sync message finished
        running = false;
        stopped = true;
        
        // remove sync watchdog thread
        try {
          removeSyncWatchDogThread(r.getSyncId());
        } catch(MoaException e) {
          logger.warn("Error removing sync watchdog thread (" + r.getSyncId() + ")", e);
        }
        
        if (syncStatus.getStatus() == SyncStatus.STATUS_UNKNOWN) {
          return MoaResponse.getPartialSuccess( r.getAppDeliveryId(), String.valueOf(r.getId()));
        } else if (syncStatus.getStatus() == SyncStatus.STATUS_SUCCESS) {
          return MoaResponse.getSuccess( r.getAppDeliveryId(), String.valueOf(r.getId()));
        } else {
          return MoaResponse.getError(r.getAppDeliveryId(), String.valueOf(r.getId()),
              syncStatus.getCode(),
              syncStatus.getText(),
              syncStatus.getDocumentReference());
        }

      }
      
      if (partial) {
          return MoaResponse.getPartialSuccess( r.getAppDeliveryId(), String.valueOf(r.getId()));
      }

      return MoaResponse.getSuccess( r.getAppDeliveryId(), String.valueOf(r.getId()));
    }


    /**
    * Methode die den ZuseKopf nach einem Zustellserver (evtl mit Zertifikat) fragt.
    * Bei bedarf wird auch noch das SZR verwendet um eine Bereichsfremde BPK umzurechnen
     * @param rcpt Empfaenger
     * @param encryption notwendig (true/false)
     * @param senderprofile 
     * @param mimeTypes Liste der Mimetypes der Dokumente, die zugestellt werden sollen
     * @return Zuseserver
     * @throws MoaException
     */
    public Zuseserver getZuseserver(Rcpt rcpt, boolean encryption, boolean privateDelivery, String senderprofile, List mimeTypes) throws MoaException {

      /* SZR anfrage */
      // Falls eine bPK verwendet wurde
      if (rcpt.getPhysicalPerson() == true &&
          (rcpt.getZusebpk() == null || rcpt.getZusebpk().equals("")) &&
          rcpt.getIdentification() != null) {
        for (Iterator i = rcpt.getIdentification().iterator(); i.hasNext(); ) {
          Identification ident = (Identification)i.next();
          if (ident.getIdType().trim().startsWith("urn:publicid:gv.at:")) {
            MoaSZR mszr = new MoaSZR();
            logger.debug("found bpk ("+ident.getValue()+") ...");
            String result ="";
            result = mszr.transform(rcpt, ident.getIdType(), ident.getValue());
            rcpt.setZusebpk(result);
            break;
          }
        }
      }
      
      ZkopfRequest zr = new ZkopfRequest(mimeTypes);
      zr.setRequiresEncryption(encryption);
      // Setzen ob der Request fr den Behrdlichen oder fr einen Privaten ist
      zr.setPrivateDelivery(privateDelivery);
      
      if (rcpt.getPhysicalPerson()==true) {
        if (rcpt.getZusebpk()!=null && rcpt.getZusebpk()!="") {
          zr.queryPerson(rcpt.getZusebpk());
        } else {

                String gebdate = "";
                if (rcpt.getDateofbirth()!=null) {

                    gebdate = new SimpleDateFormat("yyyy-MM-dd").format(rcpt.getDateofbirth());
                }

                Set ia = rcpt.getInternetaddress();
                if (ia!=null) {
                    for (Iterator i=ia.iterator(); i.hasNext(); ) {

                        Internetaddress addr = (Internetaddress)i.next();
                        zr.queryPerson( rcpt.getFamilyname(), rcpt.getGivenname(), addr.getValue(), null, gebdate);
                    }
                }
                Set pa = rcpt.getPostaladdress();

                if (pa!=null) {
                    for (Iterator i=pa.iterator(); i.hasNext(); ) {

                        Postaladdress addr = (Postaladdress)i.next();
                        String street = addr.getStreetname();
                        if (addr.getBuilding() != null && !addr.getBuilding().equals("")) {
                            street += " "+addr.getBuilding();
                        }
                        if (addr.getUnit() != null && !addr.getUnit().equals("")) {
                            street += "/"+addr.getUnit();
                        }
                        if (addr.getDoor() != null && !addr.getDoor().equals("")) {
                            street += "/"+addr.getDoor();
                        }
                        zr.queryPerson( rcpt.getFamilyname(), rcpt.getGivenname(), street, addr.getPostalcode(), 
                            addr.getMunicipality(), addr.getCountrycode(), gebdate);
                    }
                }
            }
        } else {
            if (rcpt.getIdentification()!=null) {
                for (Iterator i = rcpt.getIdentification().iterator(); i.hasNext(); ) {
                    Identification ident = (Identification)i.next();
                    if (ident.getIdType().trim().startsWith("urn:publicid:gv.at:")) {
                        rcpt.setZusebpk(ident.getValue()); 
                    }
                }
            }
            if (rcpt.getZusebpk()!=null && !rcpt.getZusebpk().equals("")) {
                logger.debug("Company mit ID");
                zr.queryCompany(rcpt.getZusebpk());            
            } else {
                logger.debug("Company mit Adresse");
                Set pa = rcpt.getPostaladdress();
                if (pa != null) {
                    for (Iterator i=pa.iterator(); i.hasNext(); ) {
                        Postaladdress addr = (Postaladdress)i.next();
                        String street = addr.getStreetname();
                        if (addr.getBuilding() != null && !addr.getBuilding().equals("")) {
                            street += " "+addr.getBuilding();
                        }
                        if (addr.getUnit() != null && !addr.getUnit().equals("")) {
                            street += "/"+addr.getUnit();
                        }
                        if (addr.getDoor() != null && !addr.getDoor().equals("")) {
                            street += "/"+addr.getDoor();
                        }

                        zr.queryCompany( rcpt.getFullname(), street, addr.getPostalcode(), addr.getMunicipality(), addr.getCountrycode());
                    }
                }
            }
        }
      	Zuseserver zs = zr.getZs();
        if (zs != null) {
            rcpt.setZusebpk(zr.getZs().getBpk());
        } else {
            MoaConfig mcfg = new MoaConfig();
            Properties prob = mcfg.getSenderInfo(senderprofile);

            if (prob.getProperty("DefaultURL")!=null) {
                zs = new Zuseserver();
                zs.setUrl((String)prob.getProperty("DefaultURL"));
                zs.setBpk((String)prob.getProperty("DefaultBPK"));
                zs.setMimetypes((String)prob.getProperty("DefaultMIME"));
                rcpt.setZusebpk( zs.getBpk());
            }
	    }
        return zs;
    }

}
