package moazs.external;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.rpc.Call;
import javax.xml.rpc.Service;
import javax.xml.rpc.ServiceException;
import javax.xml.rpc.ServiceFactory;
import javax.xml.transform.TransformerException;

import moazs.Constants;
import moazs.api.MoaException;
import moazs.util.MoaConfig;
import moazs.util.MoaXMLUtils;
import moazs.util.MoaXPath;

import org.apache.axis.message.SOAPBodyElement;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import at.gv.egovernment.moa.util.DOMUtils;

public class MoaSP {

  static private Logger logger = Logger.getLogger(MoaSP.class);
  
  // moa-sp endpoint url
  private String endpoint = null;
  private String trustProfileId = null;
  
  // namespaces
  private String NS_MOA = null;
  private String NS_DSIG = null;
  private String NS_ZUS = null;

  private MoaConfig mcfg;
  
  public MoaSP() {
    this.mcfg = new MoaConfig();
    this.endpoint = mcfg.getProperty("endpoint", "moasp");
    this.trustProfileId = mcfg.getDeliveryNotificationProperty("trustprofile");
    this.NS_DSIG = mcfg.getProperty("dsig", "Namespaces");
    this.NS_MOA = mcfg.getProperty("msp", "Namespaces");
    this.NS_ZUS = mcfg.getProperty("zus", "Namespaces");
  }
  
  public boolean verify(Element signatureEnvironment, String signatureLocation) throws MoaException {
    
    if (this.endpoint == null) {
      logger.error("Please specify the endpoint for MOA-SP first.");
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "No endpoint for MOA-SP service specified.");
    }
    
    if (this.trustProfileId == null) {
      logger.error("Please specify the trustprofile for MOA-SP first.");
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "No trustprofile for MOA-SP service specified.");
    }
    
    try {
      logger.debug("Creating MOA-SP request.");
      Element request = makeRequest(signatureEnvironment, signatureLocation, this.trustProfileId);
      logger.debug("MOA-SP request created successfully.");
      if (mcfg.getDebugRequests()) {
        logger.debug("MOA-SP request: " + MoaXMLUtils.serializeToString(request, true));
      }
      logger.debug("Calling MOA-SP");
      Service service = ServiceFactory.newInstance().createService(new QName("VerifyXMLSignatureRequest"));
      Call call = (Call) service.createCall();
      SOAPBodyElement body = new SOAPBodyElement(request);
      SOAPBodyElement[] params = new SOAPBodyElement[] { body };
  
      call.setTargetEndpointAddress(this.endpoint);
      logger.debug("Connecting to MOA-SP: " + this.endpoint);
      logger.debug("MOA-SP Trustprofile: " + this.trustProfileId);
      Vector responses = (Vector) call.invoke(params);
  
      SOAPBodyElement response = (SOAPBodyElement) responses.get(0);
      Document resdoc = null;
      try {
          resdoc = response.getAsDocument();
          if (mcfg.getDebugRequests()) {
            logger.debug("MOA-SP response: " + MoaXMLUtils.serializeToString(resdoc.getDocumentElement(), true));
          }
      } catch (Exception e) {
        logger.warn("Error while communicating with MOA-SP.", e);
        throw new MoaException("Communication error with MOA-SP");
      }
      
      // evaluate response codes
      // xmldsig check code
      String xmlDsigCodeString = MoaXPath.getFirstString(resdoc.getDocumentElement(), "msp:SignatureCheck/msp:Code/text()");
      int xmlDsigCode = Integer.parseInt(xmlDsigCodeString);
      
      // xmldsig check manifest code
      int xmlDsigManifestCode = 0;
      String xmlDsigManifestCodeString = MoaXPath.getFirstString(resdoc.getDocumentElement(), "msp:XMLDSIGManifestCheck/msp:Code/text()");
      if (xmlDsigManifestCodeString != null) {
        xmlDsigManifestCode = Integer.parseInt(xmlDsigManifestCodeString);
      }
      
      // certificate check code
      String certificateCodeString = MoaXPath.getFirstString(resdoc.getDocumentElement(), "msp:CertificateCheck/msp:Code/text()");
      int certificateCode = Integer.parseInt(certificateCodeString);
      
      logger.debug("Signature Verification result: Signature Check(" + xmlDsigCode + ") XMLDSIG Manifest Check(" + xmlDsigManifestCode + ") Certificate Check(" + certificateCode +")");
      
      boolean isSignatureValid = (xmlDsigCode == 0) &&
                      (xmlDsigManifestCode == 0) &&
                      (certificateCode == 0);
      
      logger.debug("Signature valid: " + isSignatureValid);
      
      return isSignatureValid;
    } catch(ServiceException e) {
      logger.error(e);
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error connecting to MOA-SP service.");
    } catch(RemoteException e) {
      logger.error(e);
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error connecting to MOA-SP service.");
    } catch(IOException e) {
      logger.error(e);
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error connecting to MOA-SP service.");
    }
  }
  
  private Element makeRequest(Element env, String loc, String trustProfileId) throws MoaException {
    try {
      DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
      dbf.setNamespaceAware(true);
      DocumentBuilder db = dbf.newDocumentBuilder();
      Document result = db.getDOMImplementation().createDocument(NS_MOA, "moa:VerifyXMLSignatureRequest", null);
      Element root = result.getDocumentElement();
      
      root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:moa", NS_MOA);
      root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:dsig", NS_DSIG);
      root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:zus", NS_ZUS);
      
      // verify signature info
      Element vsi = result.createElementNS(NS_MOA, "moa:VerifySignatureInfo");
      root.appendChild(vsi);
      
      // verify signature environment
      Element vse = result.createElementNS(NS_MOA, "moa:VerifySignatureEnvironment");
      vsi.appendChild(vse);
      
      // xml content element
      Element xmlcontent = result.createElementNS(NS_MOA, "moa:XMLContent");
      vse.appendChild(xmlcontent);
      
      // add content
      xmlcontent.appendChild(result.importNode(env, true));

      // signature location
      Element vsl = result.createElementNS(NS_MOA, "moa:VerifySignatureLocation");
      vsl.setAttribute("xmlns:zus", NS_ZUS);
      vsl.setAttribute("xmlns:dsig", NS_DSIG);
      vsi.appendChild(vsl);
      vsl.appendChild(result.createTextNode(loc));
      
      Element tpid = result.createElementNS(NS_MOA, "moa:TrustProfileID");
      tpid.appendChild(result.createTextNode(trustProfileId));
      root.appendChild(tpid);
      
      return root;
    } catch(ParserConfigurationException e) {
      logger.warn("Error while creating MOA-SP request.", e);
      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
    }

  }
  
    
}