/*
 * Copyright 2004 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package moazs.external;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.rpc.Call;
import javax.xml.rpc.Service;
import javax.xml.rpc.ServiceFactory;

import moazs.Constants;
import moazs.api.MoaException;
import moazs.modell.Payload;
import moazs.modell.XmlDocument;
import moazs.util.Base64;
import moazs.util.MoaConfig;
import moazs.util.MoaXPath;

import org.apache.axis.message.SOAPBodyElement;
import org.apache.log4j.Logger;
import org.apache.xml.serialize.XMLSerializer;
import org.hibernate.classic.Session;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Klasse zum signieren von Zustellstcken via MoaSPSS
 */

public class MoaSS {
    private String NS_MOA = null;
    private String NS_DSIG = null;

    static private Logger logger = Logger.getLogger(MoaSS.class);
    private String endpoint = null;
    private String keystorepw = null;

    /**
     * Standard Konstruktor
     */
    public MoaSS() {
        MoaConfig mcfg = new MoaConfig();
        this.endpoint = mcfg.getProperty("endpoint", "moass");

        this.NS_DSIG = mcfg.getProperty("dsig", "Namespaces");
        this.NS_MOA = mcfg.getProperty("mss", "Namespaces");
    }

    /**
     * Anfrage an MoaSS
     * 
     * @param deckblatt
     *            XmlDocument
     * @param payload
     *            Set von Payloads
     * @param key
     *            Schluessel
     * @return bytearray
     * @throws MoaException
     */
    public byte[] sign(XmlDocument deckblatt, Set payload, String key, Session sess) throws MoaException {
        byte[] result = null;
        try {
            long start = System.currentTimeMillis();
            Document request = makeRequest(deckblatt, payload, key);
            long stop = System.currentTimeMillis();
            logger.debug( "SIGNREQUEST: " + (stop - start));

            Service service = ServiceFactory.newInstance().createService(new QName("CreateXMLSignatureRequest"));
            Call call = (Call) service.createCall();
            SOAPBodyElement body = new SOAPBodyElement(request.getDocumentElement());
            SOAPBodyElement[] params = new SOAPBodyElement[] { body };

            call.setTargetEndpointAddress(this.endpoint);
            Vector responses = (Vector) call.invoke(params);

            SOAPBodyElement response = (SOAPBodyElement) responses.get(0);
            Document resdoc = null;
            try {
                resdoc = response.getAsDocument();
            } catch (Exception e) {
                throw new MoaException("komunikationsfehler mit MoaSS");
            }

            Element err = MoaXPath.getFirstElement(resdoc.getDocumentElement(),
                    "/mss:CreateXMLSignatureResponse/mss:ErrorResponse");

            if (err != null) {
                String code = MoaXPath.getFirstString(err, "mss:ErrorCode/text()");
                String msg = MoaXPath.getFirstString(err, "mss:Info/text()");

                logger.info("Error occured signing document with moa-ss: " + code + " " + msg);
                throw new MoaException(Integer.parseInt("2" + code), msg);
            }

            Element rese = MoaXPath.getFirstElement(resdoc.getDocumentElement(),
                    "/mss:CreateXMLSignatureResponse/mss:SignatureEnvironment");

            XMLSerializer serializer = new XMLSerializer();
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            serializer.setOutputByteStream(bos);

            DocumentFragment df = resdoc.createDocumentFragment();
            Node tmp = rese.getFirstChild();
            while (tmp != null && tmp.getNodeName().charAt(0) == '#') {
                tmp = tmp.getNextSibling();
            }
            if (tmp != null) {
        		df.appendChild( resdoc.createProcessingInstruction("xml-stylesheet", "href=\""+deckblatt.getPstylesheetName()+"\" type=\"text/xsl\""));
                df.appendChild(tmp);
                serializer.serialize(df);
                result = bos.toByteArray();
            }

        } catch (java.rmi.RemoteException re) {
            logger.warn( "RemoteException", re);
        } catch (java.io.IOException ioe) {
            logger.warn( "IO Exception", ioe);
        } catch (javax.xml.rpc.ServiceException se) {
            logger.warn( "ServiceException", se);
        }

        return result;
    }

    /**
     * Erstellen des Requests fuer Anfrage an MoaSS
     * 
     * @param deckblatt
     *            XmlDocument
     * @param payload
     *            Set
     * @param key
     *            Schluessel
     * @return DOM-Document
     * @throws MoaException
     */
    public Document makeRequest(XmlDocument deckblatt, Set payload, String key) throws MoaException {
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document result = db.getDOMImplementation().createDocument(NS_MOA, "moa:CreateXMLSignatureRequest", null);
            Element root = result.getDocumentElement();
            root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:moa", NS_MOA);
            root.setAttributeNS("http://www.w3.org/2000/xmlns/", "xmlns:dsig", NS_DSIG);

            Element keyidentifier = result.createElementNS(NS_MOA, "moa:KeyIdentifier");
            keyidentifier.appendChild(result.createTextNode(key));

            Element signatureinfo = result.createElementNS(NS_MOA, "moa:SingleSignatureInfo");
            signatureinfo.setAttributeNS(NS_MOA, "SecurityLayerConformity", "false");
            root.appendChild(keyidentifier);
            root.appendChild(signatureinfo);

            //  DataObjectInfo fuer Deckblatt
            Element doix = result.createElementNS(NS_MOA, "moa:DataObjectInfo");
            doix.setAttributeNS(NS_MOA, "Structure", "detached");
            Element dataobjectx = result.createElementNS(NS_MOA, "moa:DataObject");
            dataobjectx.setAttributeNS(NS_MOA, "Reference", "");

            Element ctipx = result.createElementNS(NS_MOA, "moa:CreateTransformsInfoProfile");
            Element ctix = result.createElementNS(NS_MOA, "moa:CreateTransformsInfo");
            Element dts = result.createElementNS(NS_DSIG, "dsig:Transforms");
            Element dt = result.createElementNS(NS_DSIG, "dsig:Transform");
            dt.setAttributeNS(NS_DSIG, "Algorithm", "http://www.w3.org/2000/09/xmldsig#enveloped-signature");
            dts.appendChild(dt);

            if (deckblatt.getSstylesheet() != null) {
                Element dt2 = result.createElementNS(NS_DSIG, "dsig:Transform");
                dt2.setAttributeNS(NS_DSIG, "Algorithm", "http://www.w3.org/TR/1999/REC-xslt-19991116");
                Document stylesheet = null;
                try {
                    ByteArrayInputStream bis = new ByteArrayInputStream(deckblatt.getSStyleSheetAsByteArray());

                    // Parser/DOMBuilder instanzieren
                    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                    factory.setNamespaceAware(true);
                    DocumentBuilder builder = factory.newDocumentBuilder();

                    // XML Datei in einen DOM-Baum umwandeln
                    stylesheet = builder.parse(bis);

                } catch (org.xml.sax.SAXException se) {
                    logger.warn( "SAXException", se);
                    throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
                } catch (java.io.IOException ioe) {
                    logger.warn( "IOException", ioe);
                    throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
                }
                Node xslt = result.importNode(stylesheet.getDocumentElement(), true);
                dt2.appendChild(xslt);
                dts.appendChild(dt2);
            }
            Element fdmx = result.createElementNS(NS_MOA, "moa:FinalDataMetaInfo");
            Element mimex = result.createElementNS(NS_MOA, "moa:MimeType");
            mimex.appendChild(result.createTextNode(deckblatt.getResultingMimetype()));

            signatureinfo.appendChild(doix);
            doix.appendChild(dataobjectx);

            doix.appendChild(ctipx);
            ctipx.appendChild(ctix);
            ctix.appendChild(dts);
            ctix.appendChild(fdmx);
            fdmx.appendChild(mimex);

            // DataObjektInfo fuer alle anderen Payloadeintraege
            if (payload != null) {
                for (Iterator i = payload.iterator(); i.hasNext();) {
                    Payload p = (Payload) i.next();
                    Element doi = result.createElementNS(NS_MOA, "moa:DataObjectInfo");
                    doi.setAttributeNS(NS_MOA, "Structure", "detached");
                    doi.setAttributeNS(NS_MOA, "ChildOfManifest", "true");
                    Element dataobject = result.createElementNS(NS_MOA, "moa:DataObject");
                    dataobject.setAttributeNS(NS_MOA, "Reference", p.getFilename());

                    //byte[] content = null;
                    String content = null;
                    Document doc = null;
                    try {
                      if (!p.getBase64()) {
                          // content = Base64.encode(p.getData());
                          content = Base64.encodeBytes(p.getDataAsByteArray());
                      } else {
                          content = new String(p.getDataAsByteArray());
                      }
                    } catch(IOException e) {
                      logger.warn("IOException...", e);
                      throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
                    }
                    Element b64content = result.createElementNS(NS_MOA, "moa:Base64Content");
                    b64content.appendChild(result.createTextNode(content));
                    dataobject.appendChild(b64content);

                    signatureinfo.appendChild(doi);
                    doi.appendChild(dataobject);

                    Element ctip = result.createElementNS(NS_MOA, "moa:CreateTransformsInfoProfile");
                    Element cti = result.createElementNS(NS_MOA, "moa:CreateTransformsInfo");

                    Element fdm = result.createElementNS(NS_MOA, "moa:FinalDataMetaInfo");
                    Element mime = result.createElementNS(NS_MOA, "moa:MimeType");
                    mime.appendChild(result.createTextNode(p.getMimetype()));
                    doi.appendChild(ctip);
                    ctip.appendChild(cti);
                    cti.appendChild(fdm);
                    fdm.appendChild(mime);
                }
            }
            // CreateSignatureInfo
            Element csi = result.createElementNS(NS_MOA, "moa:CreateSignatureInfo");
            Element cse = result.createElementNS(NS_MOA, "moa:CreateSignatureEnvironment");
            Element xmlcontent = result.createElementNS(NS_MOA, "moa:XMLContent");
            Document doc = null;
            try {
                ByteArrayInputStream bis = new ByteArrayInputStream(deckblatt.getDataAsByteArray());

                // Parser/DOMBuilder instanzieren
                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                factory.setNamespaceAware(true);
                DocumentBuilder builder = factory.newDocumentBuilder();

                // XML Datei in einen DOM-Baum umwandeln
                doc = builder.parse(bis);

            } catch (org.xml.sax.SAXException se) {
                logger.warn( "SAXException", se);
                throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
            } catch (java.io.IOException ioe) {
                logger.warn( "IOException", ioe);
                throw new MoaException(Constants.ERROR_PROCESSING_UNCLASSIFIED, "Unclassified error while handling request.");
            }
            if (doc == null) {
                throw new MoaException("Invalid XMLContent");
            }
            Node xx = result.importNode(doc.getDocumentElement(), true);
            xmlcontent.appendChild(xx);

            Element csep = result.createElementNS(NS_MOA, "moa:CreateSignatureEnvironmentProfile");
            Element csl = result.createElementNS(NS_MOA, "moa:CreateSignatureLocation");
            csl.appendChild(result.createTextNode(deckblatt.getXpath()));
            csl.setAttributeNS(NS_MOA, "Index", Integer.toString(deckblatt.getXpathIndex()));

            signatureinfo.appendChild(csi);
            csi.appendChild(cse);
            cse.appendChild(xmlcontent);

            csi.appendChild(csep);
            csep.appendChild(csl);

            return result;
        } catch (javax.xml.parsers.ParserConfigurationException pce) {
            logger.warn( "ParserConfigException", pce);
            throw new MoaException(pce);
        }
    }
}
