/*
 * Copyright 2004 Bundeskanzleramt sterreich
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package moazs.external;

import iaik.utils.Util;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.transform.TransformerException;

import moazs.Constants;
import moazs.api.MoaException;
import moazs.modell.Zuseserver;
import moazs.util.MoaConfig;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.log4j.Logger;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import at.gv.egovernment.moa.util.DOMUtils;

/**
 * Klasse zur Abfrage von Zustellinformationen fr natrliche oder nicht natrliche Personen vom ZUSEKopf
 * <p>
 * Rueckgabe der Daten erfogt ueber HashMap getZuseInfo()
 * 
 */
public class ZkopfRequest {

  private static Logger log = Logger.getLogger(ZkopfRequest.class);
  
//    private static final String ERROR = "Error";
//    private static final String SUCCESS = "Success";
//    private static final String SERVER = "Server";
    private static final String GVZBPK = "gvZbPK";
    private static final String ZUSEURLID = "ZUSEUrlID";
    private static final String MIMETYPES = "MIMETypes";
    private static final String X509 = "X509";
    private static final String EDID = "edID";
    private static final String SERVER = "Server";
    
    private boolean requiresEncryption = false;

    private Zuseserver zs = null;
    private MoaConfig mcfg = null;

    private HashMap zuseInfo = new HashMap();
    private LinkedList withX509 = new LinkedList();
    private LinkedList noX509 = new LinkedList();

    // list of document mimetypes that should be delivered
    // all mimetypes have to be lower case
    private List inputMimeTypes = new ArrayList();
    
    private boolean checkMimeTypes = true;
    
    private boolean privateDelivery = false;

    /**
     * constructor
     */
    public ZkopfRequest(List mimeTypes) {
        super();
        mcfg = new MoaConfig();
        this.inputMimeTypes = mimeTypes;
        if (inputMimeTypes != null) {
          String mimeTypesString = "";
          for (int i=0; i<inputMimeTypes.size(); i++) {
            mimeTypesString += inputMimeTypes.get(i) + " ";
          }
          log.debug("Mimetypes to check against: " + inputMimeTypes);
        } else {
          log.debug("No mimetypes to check against.");
        }
        // check if we have to ignore mime type check
        this.checkMimeTypes = mcfg.getCheckMimetypes();
    }

    /**
     * Anfrage zu einer natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mit Name + Verstndigungsadresse + Geburtsdatum
     * 
     * @param nachName
     *            Nachname der natrlichen Person
     * @param vorName
     *            Vorname der natrlichen Person
     * @param mailadr
     *            Mailadress
     * @param telefon
     *            Telefonnummer
     * @param geburtstag
     *            Geburtsdatum
     * @throws MoaException
     */
    public void queryPerson(String nachName, String vorName, String mailadr, String telefon, String geburtstag)
            throws MoaException {
        NameValuePair firstName = new NameValuePair("sn", nachName);
        NameValuePair lastName = new NameValuePair("givenName", vorName);
        NameValuePair email = new NameValuePair("mail", mailadr);
        NameValuePair phone = new NameValuePair("tel", telefon);
        NameValuePair birthday = new NameValuePair("gvBirthdate", geburtstag);
        queryZKOPF(new NameValuePair[] { firstName, lastName, email, phone, birthday });
    }

    /**
     * Anfrage zu einer natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mit Name + ZMR konformer Adresse + [Geburtsdatum fr RSA Zustellung]
     * 
     * @param nachName
     *            Nachname
     * @param vorName
     *            Vorname
     * @param strasse
     *            Strasse
     * @param plz
     *            Postleitzahl
     * @param ort
     *            Wohnort
     * @param landesCode
     *            Landescode (e.g.AT)
     * @param geburtstag
     *            Geburtsdatum
     * @throws MoaException
     */
    public void queryPerson(String nachName, String vorName, String strasse, String plz, String ort, String landesCode,
            String geburtstag) throws MoaException {

        NameValuePair firstName = new NameValuePair("sn", nachName);
        NameValuePair lastName = new NameValuePair("givenName", vorName);
        NameValuePair street = new NameValuePair("StreetName", strasse);
        NameValuePair zip = new NameValuePair("postalCode", plz);
        NameValuePair city = new NameValuePair("Municipality", ort);
        NameValuePair cc = new NameValuePair("CountryCode", landesCode);
        NameValuePair birthday = new NameValuePair("gvBirthdate", geburtstag);

        queryZKOPF(new NameValuePair[] { firstName, lastName, street, zip, city, cc, birthday });
    }

    /**
     * Anfrage zu einer natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mittels verschlsselter oder unverschlsselter Zustell-bPK
     * 
     * @param zustellBPK
     *            zur Anfrage beim ZUSEKopf
     * @throws MoaException
     */
    public void queryPerson(String zustellBPK) throws MoaException {
        NameValuePair zbPK = new NameValuePair("gvZbPK", zustellBPK);
        queryZKOPF(new NameValuePair[] { zbPK });
    }

    /**
     * Anfrage zu einer nicht natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mittels Stammzahl
     * 
     * @param stammzahl
     *            Stammzahl aus dem Firmenregister
     * @throws MoaException
     */
    public void queryCompany(String stammzahl) throws MoaException {
      NameValuePair rootnr = new NameValuePair("gvSourcePin", stammzahl);
      queryZKOPF(new NameValuePair[] { rootnr });
    }

    /**
     * Anfrage zu einer nicht natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mit Bezeichnung + ZMR konformer Adresse
     * 
     * @param name
     *            Firmenname
     * @param strasse
     *            Strasse
     * @param plz
     *            Postleitzahl
     * @param ort
     *            Ort an welchem die Firma ansssig ist
     * @throws MoaException
     */
    public void queryCompany(String name, String strasse, String plz, String ort, String landesCode)
            throws MoaException {
        NameValuePair company = new NameValuePair("cn", name);
        NameValuePair street = new NameValuePair("StreetName", strasse);
        NameValuePair zip = new NameValuePair("postalCode", plz);
        NameValuePair city = new NameValuePair("Municipality", ort);
        NameValuePair cc = new NameValuePair("CountryCode", landesCode);
        queryZKOPF(new NameValuePair[] { company, street, zip, city, cc });
    }

    /**
     * Anfrage zu einer nicht natuerlichen Person an den ZUSEKopf
     * <p>
     * Anfrage mit Bezeichnung + Verstndigungsadresse
     * 
     * @param name
     *            Firmenname
     * @param mailadr
     *            Strasse
     * @param telefon
     *            Telefonnummer
     * @throws MoaException
     */
    /*
     * public void queryCompany(String name, String mailadr, String telefon) { NameValuePair company = new
     * NameValuePair("cn", name); NameValuePair email = new NameValuePair("mail", mailadr); NameValuePair phone = new
     * NameValuePair("tel", telefon); queryZUSE(new NameValuePair[] { company, email, phone }); }
     */

    /**
     * Abfrage an den ZUSEKopf
     * <p>
     * Erhlt die notwendigen Wertepaare in URL kodierter Form durch ein Array von NameValuePairs
     * 
     * @param nvp
     *            Array von NameValuePairs
     * @throws MoaException
     */
    private void queryZKOPF(NameValuePair[] nvp) throws MoaException {
        String zuseURL;
        int statuscode;
        InputStream in = null;

        HttpClient client = new HttpClient();

        zuseURL = mcfg.getProperty("ZUSEurl");

        log.debug("Connecting to ZuseKopf: " + zuseURL);
        
        GetMethod gm = new GetMethod(zuseURL);
        
        // extend query string by private delivery flag
        if (this.privateDelivery) {
          NameValuePair[] newValues = new NameValuePair[nvp.length + 1];
          System.arraycopy(nvp, 0, newValues, 0, nvp.length);
          newValues[nvp.length] = new NameValuePair("private", "true");
          nvp = newValues;
        }
        
        gm.setQueryString(nvp);
        
        if (mcfg.getDebugRequests()) {
          log.debug("ZKopf request:");
          log.debug(gm.getQueryString());
        }

        try {
            statuscode = client.executeMethod(gm);
            if (statuscode != 200) {
              log.debug("Status message: " + gm.getResponseBodyAsString());
              throw new MoaException(Constants.ERROR_ZKOPF_WRONG_STATUS_CODE, "Received status code " + statuscode + " from delivery head.");
            }
            if (mcfg.getDebugRequests()) {
              log.debug(gm.getResponseBodyAsString());
            }
            in = gm.getResponseBodyAsStream();
            Document mydoc = parseResult(in);
            if (mydoc != null) {
                //no document found within return
                //or not well formed document
                fillData(mydoc);
                in.close();
            } else {
              in.close();
              throw new IOException("No document found in delivery head response.");
            }
            
        } catch (IOException e) {
          log.warn("Error reading delivery head response", e);
          throw new MoaException(Constants.ERROR_ZKOPF_READING_RESPONSE, "Error reading delivery head response");
        }
    }

    /**
     * Methode zum parsen der Rckgabe vom ZUSEKopf
     * 
     * @param in
     *            InputStream der glesenen Daten.
     */
    private Document parseResult(InputStream in) {
        InputSource is = new InputSource(in);

        DOMParser parser = new DOMParser();
        try {
            parser.parse(is);
            Document doc = parser.getDocument();
            return doc;
        } catch (SAXException se) {
            log.warn("Fehler bei parsen der Antwort vom ZuseKopf: ", se);
        } catch (IOException ie) {
            log.warn("Fehler beim lesen der Antwort vom Zusekopf", ie);
        }
        return null;
    }

    /**
     * Methode zum ablegen der erhaltenen Rckgabewerte
     * 
     * @param doc
     *            das aus der Antwort vom ZuseKopf gebildete DOM Dokument
     * @throws MoaException
     */
    private void fillData(Document doc) throws MoaException {
      try {
        Element errorNode = (Element) XPathAPI.selectSingleNode(doc, "/zkopf:StdAnswer/zkopf:Error", Constants.NAMESPACE_NODE);
        Element successNode = (Element) XPathAPI.selectSingleNode(doc, "/zkopf:StdAnswer/zkopf:Success", Constants.NAMESPACE_NODE);
        // evaluate error element
        if (errorNode != null) {
            String errCode = DOMUtils.getText(errorNode);
            if (errCode.compareTo("403") == 0) {
              throw new MoaException(Constants.ERROR_ZKOPF_TOO_MANY, "Person unaddressable - too many hits");
            } else if ("404".equals(errCode)) {
              throw new MoaException(Constants.ERROR_ZKOPF_NOT_FOUND, "Person unaddressable - no match found");
            } else {
              throw new MoaException(Constants.ERROR_ZKOPF_UNCLASSIFIED, "Unclassified error querying delivery head");
            }
         //ok, we have success
         } else if (successNode != null) {
        	 
        	 
        	 if (privateDelivery) {
        		 fillDataPrivateZustellung(doc);
        	 }
        	 else {
        	 
        	 
	           String gvZbPK = DOMUtils.getText(XPathAPI.selectSingleNode(successNode, "zkopf:gvZbPK", Constants.NAMESPACE_NODE));
	           //add the gvZbPK
	            if (!zuseInfo.isEmpty()) {
	                
	                if (!gvZbPK.equals((String)zuseInfo.get(GVZBPK))) {
	                  throw new MoaException(Constants.ERROR_ZKOPF_TOO_MANY, "Person unaddressable - too many hits");
	                }
	            } else {
	                zuseInfo.put(GVZBPK, gvZbPK);
	            }
	            //add ZUSEUrlID and MIMEType
	            NodeList serverList = XPathAPI.selectNodeList(successNode, "zkopf:Server", Constants.NAMESPACE_NODE);
	            HashMap[] servers = new HashMap[serverList.getLength()];
	  
	            for (int i = 0; i < serverList.getLength(); i++) {
	              servers[i] = new HashMap();
	              /*Node theNode = server.item(i).getFirstChild().getNextSibling();
	                  do {
	                      String nodeName = theNode.getNodeName();
	                      if (nodeName.compareToIgnoreCase(ZUSEURLID) == 0 || nodeName.compareToIgnoreCase(MIMETYPES) == 0
	                              || nodeName.compareToIgnoreCase(X509) == 0) {
	                          String nodeValue = theNode.getFirstChild().getNodeValue();
	                          servers[i].put(nodeName, nodeValue);
	                      }
	                      theNode = theNode.getNextSibling();
	                  } while (theNode != null);*/
	              Element serverElement = (Element) serverList.item(i);
	              String zuseURLIdText = DOMUtils.getText(XPathAPI.selectSingleNode(serverElement, "zkopf:ZUSEUrlID", Constants.NAMESPACE_NODE));
	              servers[i].put(ZUSEURLID, zuseURLIdText);
	              String mimeTypesText = DOMUtils.getText(XPathAPI.selectSingleNode(serverElement, "zkopf:MIMETypes", Constants.NAMESPACE_NODE));
	              servers[i].put(MIMETYPES, mimeTypesText);
	              Element x509Node = (Element) XPathAPI.selectSingleNode(serverElement, "zkopf:X509", Constants.NAMESPACE_NODE);
	              if (x509Node != null) {
	                String x509Text = DOMUtils.getText(x509Node);
	                if (x509Text != null) {
	                  servers[i].put(X509, x509Text);
	                }
	              }
	              if (servers[i].containsKey(X509)) {
	                  withX509.addLast(servers[i]);
	              } else {
	                  noX509.addLast(servers[i]);
	              }
	              if (privateDelivery) {
	                  // private delivery edID Element
	                  String edIdText = DOMUtils.getText(XPathAPI.selectSingleNode(serverElement, "zkopf:edID", Constants.NAMESPACE_NODE));
	                  if (edIdText != null) {
	                    servers[i].put(EDID, edIdText);
	                  }
	                }
	            }
              
            }
        } else {
            log.warn("Antwort vom Zusekopf enthlt weder Error noch Success");
            throw new MoaException(Constants.ERROR_ZKOPF_READING_RESPONSE, "Response of delivery head contains invalid content.");
        }
      } catch(TransformerException e) {
        log.warn(e);
        throw new MoaException(Constants.ERROR_ZKOPF_UNCLASSIFIED, "Unclassified error querying delivery head.");
      }
    }
    
    
    /**
     * Methode zum Parsen von Erfolgreichen PrivateZustellung ZuseKopf Abfragen. 
     * Die Antwortsyntax bei der Einzelabfrage unterscheidet sich von der behrdlichen Zustellung dadurch,
     *  dass im Erfolgsfall kein einzelnes gvZbPK Element fr alle gefundenen Zustelldienste zurckgegeben wird,
     *   sondern fr jeden einzelnen Zustelldienst ein edID Element retourniert wird.
     * @param doc
     */
    private void fillDataPrivateZustellung(Document doc ) throws MoaException {
    	
        //add ZUSEUrlID and MIMEType
        NodeList server = doc.getElementsByTagName(SERVER);
        //NodeList nextNode = doc.getElementsByTagName(SERVER).item(0).getChildNodes();
        HashMap[] hm = new HashMap[server.getLength()];

        for (int i = 0; i < server.getLength(); i	++) {
            hm[i] = new HashMap();
            Node theNode = server.item(i).getFirstChild().getNextSibling();
            boolean hasEdID = false;
            do {
        
                String nodeName = theNode.getNodeName();
                if (nodeName.compareToIgnoreCase(ZUSEURLID) == 0 || nodeName.compareToIgnoreCase(MIMETYPES) == 0 || nodeName.compareToIgnoreCase(X509) == 0 || nodeName.compareToIgnoreCase(EDID)==0) {
                    String nodeValue = DOMUtils.getText(theNode);
                    log.debug("Name: " + nodeName + " Value:" + nodeValue);
                    if (nodeName.compareToIgnoreCase(EDID) == 0 &&  (nodeValue == null || nodeValue.length() == 0))
                       throw new MoaException(Constants.ERROR_ZKOPF_NOT_FOUND, "404 Person unaddressable - EDID not filled"); 	
                    hm[i].put(nodeName, nodeValue);
                }
                theNode = theNode.getNextSibling();
            } while (theNode != null);

            if (!(hm[i].containsKey(ZUSEURLID) && hm[i].containsKey(MIMETYPES)))
                throw new MoaException("404 Person unaddressable - no match found");

            if (hm[i].containsKey(X509)) {
                withX509.addLast(hm[i]);
            } else {
                noX509.addLast(hm[i]);
            }
            
          
        }
    	
    	
    	
    }
    
    

    /**
     * Methode zur berprfung der zurckgegebenen MIMETypes
     * 
     * @param ll
     *            LinkedList mit Rckgabewerten vom ZuseKopf
     * @return LinkedList eine Liste mit allen Wertepaaren, die die input mime types untersttzen
     */
    private LinkedList checkMimeTypes(LinkedList nll) {
        //HashMap[] hm = new HashMap[nll.size()];
      HashMap hm;
      LinkedList rl = new LinkedList();
      for (int i = 0; i < nll.size(); i++) {
        hm = new HashMap();
        hm.putAll((Map) nll.get(i));
        if (supportsAllMimeTypes(hm.get(MIMETYPES).toString()) || !this.checkMimeTypes) {
          rl.addLast(hm);
        }
      }
      return rl;
    }
    
    private boolean supportsAllMimeTypes(String mimeTypesString) {
      /*String[] mimeTypes = mimeTypesString.split(",");
      if (mimeTypes == null || mimeTypes.length == 0) {
        return false;
      }
      List mimeTypesList = Arrays.asList(mimeTypes);*/
      HashMap map = mimeTypesStringToHashMap(mimeTypesString);
      if (inputMimeTypes != null) {
        for (int i=0; i<inputMimeTypes.size(); i++) {
          if (inputMimeTypes.get(i) != null) {
            String mimeType = ((String) inputMimeTypes.get(i)).trim().toLowerCase();
            if (!isMimeTypeSupported(map, mimeType)) {
              return false;
            }
          }
        }
      }
      return true;
    }

    private boolean isMimeTypeSupported(HashMap map, String mimeType) {
      if (mimeType.indexOf("/") == -1) {
        return false;
      }
      String prefix = mimeType.split("/")[0];
      String postfix = mimeType.split("/")[1];
      
      // check for supported prefix
      List wildCardList = (List) map.get("*");
      List specList = (List) map.get(prefix);
      
      if (wildCardList != null) {
        if (wildCardList.contains("*") || wildCardList.contains(postfix)) {
          return true;
        }
      }
      if (specList != null) {
        if (specList.contains("*") || specList.contains(postfix)) {
          return true;
        }
      }
      return false;
    }
    
    private HashMap mimeTypesStringToHashMap(String mimeTypesString) {
      String[] mimeTypes = mimeTypesString.split(",");
      if (mimeTypes == null || mimeTypes.length == 0) {
        return new HashMap();
      }
      HashMap ret = new HashMap();
      for (int i=0; i<mimeTypes.length; i++) {
        if (mimeTypes[i].indexOf("/") == -1) {
          continue;
        }
        String prefix = mimeTypes[i].split("/")[0];
        String postfix = mimeTypes[i].split("/")[1];
        
        List entryList = (List) ret.get(prefix);
        if (entryList == null) {
          entryList = new ArrayList();
        }
        if (!entryList.contains(postfix)) {
          entryList.add(postfix);
        }
        ret.put(prefix, entryList);
      }
      return ret;
    }

    
    /**
     * Methode zum berprfen und zuflligen Auswhlen eines Zuseservers aus der Liste der ermittelten
     * ZuseKopf-Rckgabewerte
     * 
     * @throws MoaException
     */
    private void check() throws MoaException {
        
        // filter out not supported mimetypes
        if (!withX509.isEmpty()) {
          log.debug("Checking mimetypes for servers with encryption certificate.");
          withX509 = checkMimeTypes(withX509);
        }
        if (!noX509.isEmpty()) {
          log.debug("Checking mimetypes for servers without encryption certificate.");
          noX509 = checkMimeTypes(noX509);
        }

        // now check
        if (!withX509.isEmpty()) {
          log.debug("Searching for delivery server with X509 certificate.");
          if (!setPreferredZuseServer(withX509)) {
            log.debug("No preferred delivery server found. Choosing random one of (" + withX509.size() + ").");
            setRandom(withX509);
          }
        } else if (isRequiresEncryption()) {
            zuseInfo.clear();
            throw new MoaException(Constants.ERROR_ZKOPF_NO_PK_AV, "No public key available");
        } else if (noX509.isEmpty()) {
            zuseInfo.clear();
        } else {
          log.debug("Searching for delivery server without X509 certificate.");
          if (!setPreferredZuseServer(noX509)) {
            log.debug("No preferred delivery server found. Choosing random one of (" + noX509.size() + ").");
            setRandom(noX509);
          } 
        }
    }

    private boolean setPreferredZuseServer(LinkedList ll) {
      List pref = mcfg.getPreferredServers();
      for (int i=0; i<pref.size(); i++) {
        if (setSinglePreferredZuseServer((String) pref.get(i), ll)) {
          return true;
        }
      }
      return false;
    }
    
    private boolean setSinglePreferredZuseServer(String zuseURL, LinkedList ll) {
      for (int i=0; i<ll.size(); i++) {
        Map props = (Map) ll.get(i);
        String zURL = (String) props.get(ZUSEURLID);
        if (zuseURL.equals(zURL)) {
          log.debug("Found preferred delivery service: " + zuseURL);
          zuseInfo.putAll(props);
          return true;
        }
      }
      return false;
    }
    
    /**
     * Methode zur Zuflligen Auswahl eines Elementes aus der bergebenen Liste Ermitteltes Element wird in die HashMap
     * zuseInfo geschrieben.
     * 
     * @param ll
     *            Liste mit Zuseservern
     */
    public void setRandom(LinkedList ll) {
        int number = (int) (Math.random() * ll.size());
        zuseInfo.putAll((Map) ll.get(number));
    }

    /**
     * Ermittelt den ersten Nodevalue einer bergebenen Nodelist
     * 
     * @param index
     *            Index des Nodes welcher zurueckgegeben werden soll.
     * @param nl
     *            Nodelist
     * @return value String des ermittelten NodeValues
     */
    /*private String addElementValue(int index, NodeList nl) {
        String value = "";
        Node newlength = nl.item(index).getFirstChild();
        if (newlength == null) {
            //String deristleer = "leer";
        }
        int laenge = nl.getLength();
        if (laenge > 0) {
            value = nl.item(index).getFirstChild().getNodeValue().toString();
        }
        return value;
    }*/

    /**
     * Gibt das Abfrageergebnis der Einzelabfrage an den ZUSEKopf als HashMap zurck.
     * <p>
     * Fall keine Zustellinformation gefunden wurde, enthlt die Map nur den Wert Errorcode.
     * 
     * @return Returns the zuseInfo.
     */
    public HashMap getZuseResult() {

        return zuseInfo;
    }

    /**
     * Methode zum setzen des ZuseServers aus dem zuseInfo HashMap
     * 
     * @param hm
     *            HashMap mit den notwendigen Informationen zum ermitteln eines Zuseservers
     */
    private void setZuseServer(HashMap hm) {
        if (!hm.isEmpty()) {
            zs = new Zuseserver();
            if (privateDelivery) {
              zs.setBpk(hm.get(EDID).toString());
            } else {
              zs.setBpk(hm.get(GVZBPK).toString());
            }
            zs.setUrl(hm.get(ZUSEURLID).toString());
            zs.setMimetypes(hm.get(MIMETYPES).toString());
            if (hm.containsKey(X509)) {
                zs.setCert(hm.get(X509).toString().getBytes());
            }
        }
    }

    /**
     * @return Returns the requiresEncryption.
     */
    public boolean isRequiresEncryption() {
        return requiresEncryption;
    }

    /**
     * @param requiresEncryption
     *            The requiresEncryption to set.
     */
    public void setRequiresEncryption(boolean requiresEncryption) {
        this.requiresEncryption = requiresEncryption;
    }

    /**
     * Diese Funktion uebergibt einen Zuseserver
     * 
     * @return Returns the zs.
     * @throws MoaException
     */
    public Zuseserver getZs() throws MoaException {
        check();
        setZuseServer(zuseInfo);

        return zs;
    }

    /**
     * @param zs
     *            The zs to set.
     */
    public void setZs(Zuseserver zs) {
        this.zs = zs;
    }
    
    public boolean isPrivateDelivery() {
      return privateDelivery;
    }

    public void setPrivateDelivery(boolean privateDelivery) {
      this.privateDelivery = privateDelivery;
    }
    
}
