package moazs.util;

import java.util.Iterator;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.classic.Session;
import org.hibernate.tool.hbm2ddl.SchemaExport;

/**
 * @author <a href="mailto:thomas.knall@iaik.tugraz.at">Thomas Knall</a>
 */
public final class HibernateUtil {

   private static Logger log = Logger.getLogger(HibernateUtil.class);

   private static SessionFactory sessionFactory;
   private static final ThreadLocal THREAD_LOCAL = new ThreadLocal();
   private static boolean automaticSessionHandling = false;

   private static final String[] AUTOMATIC_SESSION_HANDLING_VALUES = new String[] { "jta", "thread" };
   private static final String SESSION_HANDLING_KEY = "hibernate.current_session_context_class";

   private static Configuration configuration;
   
   /*public static void initHibernate() {
      initHibernate(new Configuration().configure());
   }

   public static void initHibernate(File configFile) {
      initHibernate(new Configuration().configure(configFile));
   }*/

   public static Configuration getConfiguration() {
      return configuration;
   }

   public static void initHibernate(MoaConfig hbmConfig) {
      if (hbmConfig != null) {
         String scm = StringUtils.trimToNull(hbmConfig.getHibernateProperties().getProperty(SESSION_HANDLING_KEY));
         if (scm != null) {
           automaticSessionHandling = scm.indexOf(AUTOMATIC_SESSION_HANDLING_VALUES[0]) != -1 || scm.indexOf(AUTOMATIC_SESSION_HANDLING_VALUES[1]) != -1;
         }
         log.debug("Evaluating hibernate property \"" + SESSION_HANDLING_KEY + "\".");
         if (automaticSessionHandling) {
            log.info("Hibernate is automatically handling session context management.");
         } else {
            log.info("Hibernate is NOT automatically handling session context management. Using build-in ThreadLocal session handling.");
         }
         configuration = hbmConfig.getHibernateConfiguration();
         try {
            // Create the SessionFactory
            log.debug("Creating initial session factory...");
            sessionFactory = configuration.buildSessionFactory();
            log.debug("Initial session factory successfully created.");
         } catch (Throwable ex) {
            log.error("Initial session factory creation failed: " + ex.getMessage());
            throw new ExceptionInInitializerError(ex);
         }
      } else {
         log.warn("There was no hibernate configuration provided.");
      }
   }

   /**
    * Checks if a session factory is currently available. If necessary a new
    * session factory is created.
    * 
    * @return current (or new) session factory
    * @throws HibernateException
    *            thrown if a hibernate error occurs
    */
   public static Session getCurrentSession() {
      if (automaticSessionHandling) {
         return sessionFactory.getCurrentSession();
      }
      Session session = (Session) THREAD_LOCAL.get();
      // Open a new Session, if this Thread has none yet
      if (session == null || !session.isConnected()) {
         session = getNewSession();
      }
      return session;
   }

   public static Session getNewSession() {
      if (automaticSessionHandling) {
         log.warn("Session is being automatically handled by hibernate. Therefore this session maybe not being newly created. Use HibernateUtil.getCurrentSession() instead.");
         return sessionFactory.getCurrentSession();
      }
      Session session = (Session) THREAD_LOCAL.get();
      if (session != null) {
         log.warn("Previous session has not been closed; closing session now.");
         closeSession();
      }
      log.debug("Opening new hibernate session...");
      try {
         session = sessionFactory.openSession();
         THREAD_LOCAL.set(session);
      } catch (HibernateException hex) {
         log.error(hex.getMessage());
      }
      return session;
   }

   /**
    * Closes the current session.
    * 
    * @throws HibernateException
    *            thrown if session is already closed or a hibernate error
    *            occurs.
    */
   public static void closeSession() {
      if (automaticSessionHandling) {
         log.warn("Session is being automatically handled by hibernate. Therefore the current session cannot be closed on demand.");
         return;
      }
      log.debug("Closing current hibernate session...");
      Session session = (Session) THREAD_LOCAL.get();
      THREAD_LOCAL.set(null);
      if (session != null) {
         try {
            session.close();
         } catch (HibernateException hex) {
            log.error(hex.getMessage());
         }
      }
   }

   public static void schemaExport(MoaConfig config, boolean forceCreation) {
      if (configuration == null) {
         initHibernate(config);
      }
      schemaExport(configuration, forceCreation);
   }

   public static void schemaExport(Configuration configuration, boolean forceCreation) {
      if (configuration == null) {
         throw new NullPointerException("configuration must not be null");
      }
      SchemaExport schemaExport = new SchemaExport(configuration);
      log.debug("Dropping schema...");
      schemaExport.setHaltOnError(false);
      schemaExport.drop(false, true);
      if (schemaExport.getExceptions().size() != 0) {
         log.error("Schema could not be completely dropped.");
         Iterator exIt = schemaExport.getExceptions().iterator();
         StringBuffer buffer = new StringBuffer();
         while (exIt.hasNext()) {
            buffer.append(((Exception) exIt.next()).getMessage());
            if (exIt.hasNext()) {
               buffer.append(", ");
            }
         }
         if (!forceCreation) {
            throw new HibernateException("Schema could not be completely dropped (" + buffer.toString() + ").");
         }
      } else {
         log.debug("Schema successfully dropped.");
      }
      log.debug("Creating schema...");
      schemaExport.create(false, true);
      if (schemaExport.getExceptions().size() != 0) {
         log.error("Schema could not be completely created.");
         Iterator exIt = schemaExport.getExceptions().iterator();
         StringBuffer buffer = new StringBuffer();
         while (exIt.hasNext()) {
            buffer.append(((Exception) exIt.next()).getMessage());
            if (exIt.hasNext()) {
               buffer.append(", ");
            }
         }
         throw new HibernateException("Schema could not be completely created (" + buffer.toString() + ").");
      } else {
         log.debug("Schema successfully created.");
      }
   }

   /*public static void save(Iterable iterable) {
      log.trace("Hibernate:save");
      Transaction tx = null;
      try {
         Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
         tx = session.beginTransaction();
         Iterator it = iterable.iterator();
         while (it.hasNext()) {
            Object obj = it.next();
            if (obj != null) {
               session.save(obj);
            }
         }
         tx.commit();
      } catch (HibernateException hibernateException) {
         log.error(hibernateException.getMessage());
         if (tx != null) {
            log.debug("Initiating rollback...");
            try {
               tx.rollback();
            } catch (HibernateException hex) {
               log.error("Rollback failed.");
            }
         }
         throw new HibernateException(hibernateException);
      } finally {
         if (!automaticSessionHandling) {
            HibernateUtil.closeSession();
         }
      }
   }

   public static void save(Object obj) {
     log.trace("Hibernate:save");
     Transaction tx = null;
     try {
        Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
        tx = session.beginTransaction();
        session.save(obj);
        tx.commit();
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     } finally {
        if (!automaticSessionHandling) {
           HibernateUtil.closeSession();
        }
     }
  }

   
   // load operation: for classic session
   public static Object get(Session session, Class clazz, Serializable ser) {
     log.trace("Hibernate:load");
     Transaction tx = null;
     try {
       if (session == null || !session.isOpen()) {
        session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
       }
       tx = session.beginTransaction();
       Object ret = session.get(clazz, ser);
       tx.commit();
       return ret;
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     }
  }

   // create sql query operation: for classic session
   public static Query createSQLQuery(Session session, String str1, String str2, Class clazz) {
     log.trace("Hibernate:load");
     Transaction tx = null;
     try {
       if (session == null) {
        session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
       }
       tx = session.beginTransaction();
       Query ret = session.createSQLQuery(str1, str2, clazz);
       tx.commit();
       return ret;
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     }
  }
   
   // find operation: for classic session
   public static List find(Session session, String searchStr) {
     log.trace("Hibernate:find");
     Transaction tx = null;
     try {
       if (session == null || !session.isOpen()) {
        session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
       }
       tx = session.beginTransaction();
       List ret = session.find(searchStr);
       tx.commit();
       return ret;
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     }
  }

   // find operation: for classic session
   public static List find(Session session, String searchStr, Object arg2, Type type) {
     log.trace("Hibernate:find");
     Transaction tx = null;
     try {
       if (session == null) {
        session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
       }
       tx = session.beginTransaction();
       List ret = session.find(searchStr, arg2, type);
       tx.commit();
       return ret;
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     } finally {
        if (!automaticSessionHandling) {
           HibernateUtil.closeSession();
        }
     }
  }

   
   public static void update(Iterable iterable) {
      log.trace("Hibernate:update");
      Transaction tx = null;
      try {
         Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
         tx = session.beginTransaction();
         Iterator it = iterable.iterator();
         while (it.hasNext()) {
            Object obj = it.next();
            if (obj != null) {
               session.update(obj);
            }
         }
         tx.commit();
      } catch (HibernateException hibernateException) {
         log.error(hibernateException.getMessage());
         if (tx != null) {
            log.debug("Initiating rollback...");
            try {
               tx.rollback();
            } catch (HibernateException hex) {
               log.error("Rollback failed.");
            }
         }
         throw new HibernateException(hibernateException);
      } finally {
         if (!automaticSessionHandling) {
            HibernateUtil.closeSession();
         }
      }
   }

   public static void delete(Iterable iterable) {
      log.trace("Hibernate:delete");
      Transaction tx = null;
      try {
         Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
         tx = session.beginTransaction();
         Iterator it = iterable.iterator();
         while (it.hasNext()) {
            Object obj = it.next();
            if (obj != null) {
               session.delete(obj);
            }
         }
         tx.commit();
      } catch (HibernateException hibernateException) {
         log.error(hibernateException.getMessage());
         if (tx != null) {
            log.debug("Initiating rollback...");
            try {
               tx.rollback();
            } catch (HibernateException hex) {
               log.error("Rollback failed.");
            }
         }
         throw new HibernateException(hibernateException);
      } finally {
         if (!automaticSessionHandling) {
            HibernateUtil.closeSession();
         }
      }
   }

   public static void delete(Object obj) {
     log.trace("Hibernate:delete");
     Transaction tx = null;
     try {
        Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
        tx = session.beginTransaction();
         if (obj != null) {
            session.delete(obj);
         }
        tx.commit();
     } catch (HibernateException hibernateException) {
        log.error(hibernateException.getMessage());
        if (tx != null) {
           log.debug("Initiating rollback...");
           try {
              tx.rollback();
           } catch (HibernateException hex) {
              log.error("Rollback failed.");
           }
        }
        throw new HibernateException(hibernateException);
     } finally {
        if (!automaticSessionHandling) {
           HibernateUtil.closeSession();
        }
     }
  }

   
   public static void saveOrUpdate(Iterable iterable) {
      log.trace("Hibernate:saveOrUpdate");
      Transaction tx = null;
      try {
         Session session = automaticSessionHandling ? sessionFactory.getCurrentSession() : HibernateUtil.getNewSession();
         tx = session.beginTransaction();
         Iterator it = iterable.iterator();
         while (it.hasNext()) {
            Object obj = it.next();
            if (obj != null) {
               session.saveOrUpdate(obj);
            }
         }
         tx.commit();
      } catch (HibernateException hibernateException) {
         log.error(hibernateException.getMessage());
         if (tx != null) {
            log.debug("Initiating rollback...");
            try {
               tx.rollback();
            } catch (HibernateException hex) {
               log.error("Rollback failed.");
            }
         }
         throw new HibernateException(hibernateException);
      } finally {
         if (!automaticSessionHandling) {
            HibernateUtil.closeSession();
         }
      }
   }


   public static void saveOrUpdateDelete(Object[] saveOrUpdateObjects, Object[] deleteObjects) {
      saveOrUpdateDelete(saveOrUpdateObjects, deleteObjects);
   }*/

   private HibernateUtil() {
   }

}
